package dev.pcvolkmer.onco.datamapper.mapper;

import dev.pcvolkmer.mv64e.mtb.*;
import dev.pcvolkmer.onco.datamapper.PropertyCatalogue;
import dev.pcvolkmer.onco.datamapper.ResultSet;
import dev.pcvolkmer.onco.datamapper.datacatalogues.*;

import java.util.List;
import java.util.stream.Collectors;

import static dev.pcvolkmer.onco.datamapper.mapper.MapperUtils.getPatientReference;

/**
 * Mapper class to load and map patient data from database table 'dk_molekulargenetik'
 *
 * @author Paul-Christian Volkmer
 * @since 0.1
 */
public class MolekulargenetikToSpecimenDataMapper implements DataMapper<TumorSpecimen> {

    private final MolekulargenetikCatalogue molekulargenetikCatalogue;
    private final TherapieplanCatalogue therapieplanCatalogue;
    private final RebiopsieCatalogue rebiopsieCatalogue;
    private final ReevaluationCatalogue reevaluationCatalogue;
    private final EinzelempfehlungCatalogue einzelempfehlungCatalogue;
    private final PropertyCatalogue propertyCatalogue;

    public MolekulargenetikToSpecimenDataMapper(
            final MolekulargenetikCatalogue molekulargenetikCatalogue,
            final TherapieplanCatalogue therapieplanCatalogue,
            final RebiopsieCatalogue rebiopsieCatalogue,
            final ReevaluationCatalogue reevaluationCatalogue,
            final EinzelempfehlungCatalogue einzelempfehlungCatalogue,
            final PropertyCatalogue propertyCatalogue
    ) {
        this.molekulargenetikCatalogue = molekulargenetikCatalogue;
        this.therapieplanCatalogue = therapieplanCatalogue;
        this.rebiopsieCatalogue = rebiopsieCatalogue;
        this.reevaluationCatalogue = reevaluationCatalogue;
        this.einzelempfehlungCatalogue = einzelempfehlungCatalogue;
        this.propertyCatalogue = propertyCatalogue;
    }

    /**
     * Loads and maps a specimen using the database id
     * The result does not include a diagnosis reference!
     *
     * @param id The database id of the procedure data set
     * @return The loaded Patient data
     */
    @Override
    public TumorSpecimen getById(int id) {
        var data = molekulargenetikCatalogue.getById(id);

        var builder = TumorSpecimen.builder();
        builder
                .id(data.getString("id"))
                .patient(getPatientReference(data.getString("patient_id")))
                .type(getTumorSpecimenCoding(data.getString("materialfixierung")))
                .collection(getCollection(data))
        // TODO add diagnosis later
        ;


        return builder.build();
    }

    /**
     * Loads and maps specimens by using the referencing KPA database id
     *
     * @param kpaId            The database id of the referencing KPA procedure data set
     * @param diagnoseReferenz The reference object to the diagnosis
     * @return The loaded Patient data
     */
    public List<TumorSpecimen> getAllByKpaId(int kpaId, Reference diagnoseReferenz) {
        var therapieplanIds = therapieplanCatalogue.getByKpaId(kpaId);

        var osMolGen = therapieplanIds.stream()
                .map(einzelempfehlungCatalogue::getAllByParentId)
                .flatMap(einzelempfehlungen ->
                        einzelempfehlungen
                                .stream()
                                .map(einzelempfehlung -> einzelempfehlung.getInteger("ref_molekulargenetik"))
                )
                .collect(Collectors.toSet());

        // Addition: Rebiopsie
        osMolGen.addAll(
                therapieplanIds.stream()
                        .map(rebiopsieCatalogue::getAllByParentId)
                        .flatMap(einzelempfehlungen ->
                                einzelempfehlungen
                                        .stream()
                                        .map(einzelempfehlung -> einzelempfehlung.getInteger("ref_molekulargenetik"))
                        )
                        .collect(Collectors.toSet())
        );

        // Addition: Reevaluation
        osMolGen.addAll(
                therapieplanIds.stream()
                        .map(reevaluationCatalogue::getAllByParentId)
                        .flatMap(einzelempfehlungen ->
                                einzelempfehlungen
                                        .stream()
                                        .map(einzelempfehlung -> einzelempfehlung.getInteger("ref_molekulargenetik"))
                        )
                        .collect(Collectors.toSet())
        );

        return osMolGen.stream()
                .map(this::getById)
                .peek(it -> it.setDiagnosis(diagnoseReferenz))
                .collect(Collectors.toList());
    }

    // TODO: Kein genaues Mapping mit Formular OS.Molekulargenetik möglich - best effort
    private TumorSpecimenCoding getTumorSpecimenCoding(String value) {
        if (value == null) {
            return null;
        }

        var resultBuilder = TumorSpecimenCoding.builder()
                .system("dnpm-dip/mtb/tumor-specimen/type");

        switch (value) {
            case "2":
                resultBuilder
                        .code(TumorSpecimenCodingCode.CRYO_FROZEN)
                        .display("Cryo-frozen");
                break;
            case "3":
                resultBuilder
                        .code(TumorSpecimenCodingCode.FFPE)
                        .display("FFPE");
                break;
            default:
                resultBuilder
                        .code(TumorSpecimenCodingCode.UNKNOWN)
                        .display("Unbekannt");
                break;
        }

        return resultBuilder.build();
    }

    private Collection getCollection(ResultSet data) {
        if (data == null || data.getString("entnahmemethode") == null || data.getString("probenmaterial") == null) {
            return null;
        }

        var methodBuilder = TumorSpecimenCollectionMethodCoding.builder()
                .system("dnpm-dip/mtb/tumor-specimen/collection/method");

        switch (data.getString("entnahmemethode")) {
            case "B":
                methodBuilder
                        .code(TumorSpecimenCollectionMethodCodingCode.BIOPSY)
                        .display("Biopsie");
                break;
            case "R":
                methodBuilder
                        .code(TumorSpecimenCollectionMethodCodingCode.RESECTION)
                        .display("Resektat");
                break;
            case "LB":
                methodBuilder
                        .code(TumorSpecimenCollectionMethodCodingCode.LIQUID_BIOPSY)
                        .display("Liquid Biopsy");
                break;
            case "Z":
                methodBuilder
                        .code(TumorSpecimenCollectionMethodCodingCode.CYTOLOGY)
                        .display("Zytologie");
                break;
            case "U":
            default:
                methodBuilder
                        .code(TumorSpecimenCollectionMethodCodingCode.UNKNOWN)
                        .display("Unbekannt");
                break;
        }

        // TODO: Kein genaues Mapping mit Formular OS.Molekulargenetik möglich - best effort
        var localizationBuilder = TumorSpecimenCollectionLocalizationCoding.builder()
                .system("dnpm-dip/mtb/tumor-specimen/collection/localization");

        switch (data.getString("probenmaterial")) {
            case "T":
                localizationBuilder
                        .code(TumorSpecimenCollectionLocalizationCodingCode.PRIMARY_TUMOR)
                        .display("Primärtumor");
                break;
            case "LK":
            case "M":
            case "ITM":
            case "SM":
                localizationBuilder
                        .code(TumorSpecimenCollectionLocalizationCodingCode.METASTASIS)
                        .display("Metastase");
                break;
            default:
                localizationBuilder
                        .code(TumorSpecimenCollectionLocalizationCodingCode.UNKNOWN)
                        .display("Unbekannt");
                break;
        }

        return Collection.builder()
                .method(methodBuilder.build())
                .localization(localizationBuilder.build())
                .build();
    }

}
